from elasticsearch import Elasticsearch, RequestsHttpConnection
from elasticsearch.helpers import bulk
from datetime import datetime
import logging
import os
import boto3
import json
import wavefront_lambda
import sys
import traceback
from requests_aws4auth import AWS4Auth

HOST_ES = os.environ.get('host', 'http://localhost:9200')
LQ_INDEX = os.environ.get('index', 'a_livequery')
LQ_DOC_TYPE = os.environ.get('doc_type', 'logs')
STATUS = os.environ.get('status', 'ACTIVE')
USER = os.environ.get('ES_USERNAME', None)
PASSWORD = os.environ.get('ES_PASSWORD', None)
BATCH_SIZE = int(os.environ.get('batch_size', 1000))
LOG_LEVEL = os.environ.get('log_level', 'info').upper()
NOTIFIER_FUNCTION = os.environ.get('notifier_function', 'liveQueryStateNotifier')
ENV = os.environ.get('env', 'local')
REGION = os.environ.get('region', 'us-east-1')

# Logging configuration
def handle_exception(exc_type, exc_value, exc_traceback):
    logger.error(traceback.format_tb(exc_traceback, 10))


logger = logging.getLogger(__name__)
handler = logging.StreamHandler(stream=sys.stdout)
file_formatter=logging.Formatter("{'@timestamp':'%(asctime)s', 'level': '%(levelname)s', 'message': '%(message)s'}")
handler.setFormatter(file_formatter)
logger.addHandler(handler)
logger.setLevel(logging.getLevelName(LOG_LEVEL))
sys.excepthook = handle_exception

# Create es client
logger.info('Creating ES client...')
if USER and PASSWORD:
    es = Elasticsearch(
        [HOST_ES],
        http_auth=(USER, PASSWORD),
        port=443,
        use_ssl=True,
        ssl_assert_hostname=False
    )
else:
    es = Elasticsearch(HOST_ES)

logger.info('Creating lambda function client...')
lambda_client = boto3.client('lambda')


def main(wavefront_registry):
    current_time = int(datetime.utcnow().timestamp() * 1000)

    logger.info('Searching for queries whose timeout times are less than {}'.format(current_time))
    search_and_update_queries(current_time=current_time, wavefront_registry=wavefront_registry)


def search_and_update_queries(current_time, wavefront_registry):
    # Build query to find expired live queries
    query = build_es_query(current_time=current_time)
    logger.debug('Built ES query for expired live queries = {}'.format(query))

    # Search for queries
    page = es.search(index=LQ_INDEX, doc_type=LQ_DOC_TYPE, scroll='2m', search_type='query_then_fetch', size=BATCH_SIZE, body=query)
    logger.debug('ES returned {} results'.format(page))

    sid = page['_scroll_id']
    scroll_size = page['hits']['total']

    # Start scrolling
    while scroll_size > 0:
        logger.info('Scrolling through result pages...')

        # Get number of results that were returned in the last scroll
        scroll_size = len(page['hits']['hits'])
        if scroll_size == 0:
            logger.info('Finished scrolling.')
            break

        logger.debug('Scroll size = {}'.format(scroll_size))

        queries = page['hits']['hits']

        # Collect query ids
        query_ids = [hit['_id'] for hit in queries]

        # Expire this set of query ids
        logger.debug('Expiring queries for ids = {}'.format(query_ids))
        set_expired(query_ids, wavefront_registry)

        # Notify admins
        try:
            logger.info('Notifying live query admins.')
            logger.debug('Building notification payloads for live query requests {}'.format(queries))
            notify_payloads = []

            for query in queries:
                query_source = query["_source"]
                if query_source["notifyOnFinish"]:
                    try:
                        notify_payloads.append(build_notifier_payload(live_query_request=query_source))
                    except Exception as e:
                        logger.error(f"Failed to build notification payload for query source = {query_source}. "
                                         "Skipping notification for this query.")

            logger.info('Built notifier payloads = {}'.format(notify_payloads))

            for payload in notify_payloads:
                notify(payload=payload)
        except Exception as e:
            logger.error("Errors occurred during email notifications.")

        # Scroll and update scroll ID
        page = es.scroll(scroll_id=sid, scroll='2m')
        sid = page['_scroll_id']


def set_expired(ids, wavefront_registry):
    """ Takes list of expired queries, updates their state to TIMED_OUT and re-indexes them. """
    update_time = int(datetime.utcnow().timestamp() * 1000)

    update_query_template = {
        "_op_type": "update",
        "_index": LQ_INDEX,
        "_type": LQ_DOC_TYPE,
        "doc": {
            "status": "TIMED_OUT",
            "statusUpdateTime": int(update_time)
        }
    }

    bulk_updates = []

    for id in ids:
        temp = update_query_template.copy()
        temp["_id"] = id
        bulk_updates.append(temp)

    logger.info('Bulk updating live queries.')
    logger.debug('Updates = {}'.format(bulk_updates))

    expired_time = int(datetime.utcnow().timestamp() * 1000)
    res = bulk(es, bulk_updates)
    logger.debug('Bulk update results = {}'.format(res))

    # send expired count to wavefront
    try:
        count = len(ids)
        wavefront_registry.counter('LIVE_QUERY.expired').inc(count)

        logger.info('Sent metrics to wavefront with count of {}'.format(count))
    except Exception as e:
        logger.error("Failed to send metrics to wavefront.")


def build_es_query(current_time):
    query = {
        "sort": [
            {
                "createTime": {
                    "order": "asc"
                }
            }
        ],
        "query": {
            "bool": {
                "must": {
                    "match": {
                        "status": STATUS
                    }
                },
                "filter": {
                    "range": {"timeoutTime": {"lte": current_time}}
                }
            }
        },
        "_source": ["_id", "createdBy", "createTime", "sql", "id", "notifyOnFinish", "name"]
    }

    return query


def notify(payload):
    """Invokes notifier function"""
    logger.debug("Invoking {} for event {}".format(NOTIFIER_FUNCTION, payload))
    invoke_response = lambda_client.invoke(FunctionName=NOTIFIER_FUNCTION,
                                           InvocationType='Event',
                                           Payload=json.dumps(payload))

    logger.debug("Notify response: {}".format(invoke_response))


def build_notifier_payload(live_query_request):
    """Builds notification payload from a live query request object"""
    return {
        "admin": live_query_request['createdBy'],
        "orgKey": live_query_request['orgKey'],
        "query": live_query_request['sql'],
        "queryId": live_query_request['id'],
        "name": live_query_request['name'],
        "startTime": {
            "nano": 0,  # this doesn't affect the notifier in any way
            "epochSecond": live_query_request['createTime'] / 1000
        },
        "status": "TIMED_OUT"
    }

@wavefront_lambda.wrapper
def lambda_handler(event, context):
    main(wavefront_registry=wavefront_lambda.get_registry())
